__author__ = "Eric Mesa"
__version__ = "2.1.1"
__license__ = "GNU GPL v3.0"
__copyright__: str = "(c) 2014 - 2023 Eric Mesa"
__email__: str = "ericsbinaryworld at gmail dot com"


import logging

from rich.logging import RichHandler

log = logging.getLogger("snapintime")
log.setLevel(logging.INFO)
RICH_FORMAT = logging.Formatter("%(message)s", "[%X]")
LOG_FILE_FORMAT = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
file_handler = logging.FileHandler('snap_in_time.log')
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(LOG_FILE_FORMAT)
console_handler = RichHandler()
console_handler.setFormatter(RICH_FORMAT)
log.addHandler(file_handler)
log.addHandler(console_handler)
print(f"Using version {__version__}")



"""Test ability to create local snapshots."""

import subprocess
from datetime import datetime
from unittest import mock

from snapintime import create_local_snapshots


@mock.patch('snapintime.create_local_snapshots.datetime')
def test_correct_date_time(datetime_mock):
    datetime_mock.now.return_value = datetime(2020, 1, 1, 12, 33)
    time_to_test = create_local_snapshots.get_date_time()
    assert time_to_test == "2020-01-01-1233"


def test_iterate_configs():
    pass


@mock.patch('snapintime.create_local_snapshots.subprocess')
def test_create_snapshot(subprocess_mock):
    subprocess_done = subprocess.CompletedProcess(args="args", returncode=0, stdout="stdout")
    subprocess_mock.run.return_value = subprocess_done
    result = create_local_snapshots.create_snapshot("2020-01-01-1233", "subvol", "backuplocation")
    assert result == {"Command": "args", "Return Code": 0, "Output": "stdout"}



"""Thin out the snapshots on disk."""

import itertools
import os
import re
import subprocess
from datetime import datetime
from typing import Pattern

import snapintime.utils.date  # type: ignore
from snapintime.utils import config as config  # type: ignore

from . import log


def split_dir_hours(subvols: list, reg_ex) -> list:
    """Return a list based on matching regular expression.

    :param subvols: A list of subvolumes.
    :param reg_ex: A re object defining the regular expression to evaluate against.
    :returns: A list that has only the items that passed the regular expression.
    """
    return [subvol for subvol in subvols if reg_ex.search(subvol) is not None]


def get_subvols_by_date(directory: str, reg_ex, remote: bool = False, remote_location: str = "") -> list:
    """Return a list based on matching regular expression.

    This is meant to produce the list that will be the input for one of the culling functions.

    :param remote_location: This should be a string like user@computer or user@IPaddress
    :param remote: True if this is taking place on the remote system
    :param directory: The directory we want to grab subvols from.
    :param reg_ex: A regular expression to apply to the directory contents.
    :returns: A list of subvolumes for culling.
    """
    if remote and remote_location:
        command = f"ssh {remote_location} ls {directory}"
        results = subprocess.run(command, capture_output=True, shell=True, check=True, text=True)
        subvols = results.stdout.split('\n')
    else:
        subvols = os.listdir(path=directory)
    return [subvol for subvol in subvols if reg_ex.search(subvol) is not None]


def btrfs_del(directory: str, subvols: list, remote: bool = False, remote_location: str = "") -> list[str]:
    """Delete subvolumes in a given directory.

    :param remote_location: This should be a string like user@computer or user@IPaddress
    :param remote: If True, working on remote system
    :param directory: The directory holding the subvolumes.
    :param subvols: A list of subvolumes to delete
    :returns: A list with the commands run and the results or, if there weren't any subvolumes\
    to delete, returns a message with that information.
    """
    return_list = []
    if subvols:
        for subvol in subvols:
            if remote and remote_location:
                command = f"ssh {remote_location} /usr/sbin/btrfs sub del {directory}/{subvol}"
            else:
                command = f"/usr/sbin/btrfs sub del {directory}/{subvol}"
            try:
                log.debug(f'{remote=}')
                log.debug(f'{remote_location=}')
                log.debug(f'{command=}')
                raw_result = subprocess.run(command, capture_output=True, shell=True, check=True, text=True)
                return_text = f"Ran {raw_result.args} with a return code of {raw_result.returncode}.\n" \
                              f"Result was {str(raw_result.stdout)}"
                return_list.append(return_text)
            except subprocess.SubprocessError as e:
                error_text = f"Ran {e.args[1]} with a return code of {e.returncode}.\nResult was {str(e.stderr)}"  # type: ignore
                return_list.append(error_text)
    else:
        return_list = [f"There was either only one or no subvolumes in {directory} at that date"]
    return return_list


def generate_daily_cull_list(dir_to_cull: list) -> list:
    """Take a list of snapshots from a directory (already reduced to one day) and cull.

    This culling will produce the closest it can to 4 snapshots\
    for that day.

    For a perfect set of 24 snapshots, it should leave behind (remove from list):

    - day1-0000
    - day1-0600
    - day1-1200
    - day1-1800

    :param dir_to_cull: A list containing snapshots. Assumes another function\
    has already reduced this list to a list containing only one day's worth of\
    snapshots.
    :returns: A list containing all the subvolumes to cull.
    """
    fourths: list[Pattern[str]] = [re.compile(r'0[0-5]\d\d$'),
                                   re.compile(r'0[6-9]\d\d$|1[0-1]\d\d$'),
                                   re.compile(r'1[2-7]\d\d$'),
                                   re.compile(r'1[8-9]\d\d$|2[0-3]\d\d$')]
    fourths_list = [split_dir_hours(dir_to_cull, fourth)[1:] for fourth in fourths]
    return list(itertools.chain.from_iterable(fourths_list))


def cull_three_days_ago(configuration: dict) -> list:
    """Cull the btrfs snapshots from 3 days ago.

    Take snapshots that were taken on the third day in the past and cull to 4 (max) snapshots.

    :param configuration: The configuration file.
    :returns: A list containing the results of running the commands.
    """
    location: str = "backuplocation"
    three_days_ago: str = snapintime.utils.date.prior_date(datetime.now(), 3).strftime("%Y-%m-%d")
    three_days_ago_reg_ex = re.compile(three_days_ago)
    return_list = []
    for subvol in configuration.values():
        subvols_three_days_ago = get_subvols_by_date(subvol.get(location), three_days_ago_reg_ex)
        three_days_ago_culled = generate_daily_cull_list(subvols_three_days_ago)
        return_list.append(btrfs_del(subvol.get(location), three_days_ago_culled))
    return return_list


def generate_weekly_cull_list(dir_to_cull: list) -> list:
    """Take a list of snapshots from a directory (already reduced to one day) and cull.

        This culling will return a list with the snapshots to remove for the given day.

        For a perfect set of snapshots, (where the user has been doing one snapshot per hour and \
        doing the daily culling) it should leave behind (remove from list):

        - day1-1800

        :param dir_to_cull: A list containing snapshots. Assumes another function\
        has already reduced this list to a list containing only one week's worth of\
        snapshots.
        :returns: A list containing all the subvolumes to cull.
        """
    sorted_dir_to_cull = sorted(dir_to_cull)
    if len(sorted_dir_to_cull) != 0:
        sorted_dir_to_cull.pop()
    return sorted_dir_to_cull


def cull_seven_days_ago(configuration: dict, remote: bool = False) -> list:
    """Cull the btrfs snapshots from 7 days ago.

    :param remote: True if culling on the remote server
    :param configuration: The configuration file.
    :returns: A list containing the results of running the commands.
    """
    seven_days_ago: str = snapintime.utils.date.prior_date(datetime.now(), 7).strftime("%Y-%m-%d")
    seven_days_ago_reg_ex = re.compile(seven_days_ago)
    return_list = []
    for subvol in configuration.values():
        location: str = "remote_subvol_dir" if remote else "backuplocation"
        subvols_seven_days_ago = get_subvols_by_date(subvol.get(location), seven_days_ago_reg_ex, True)
        log.debug(f"for {subvol.get(location)} {subvols_seven_days_ago=}")
        if remote:
            subvols_seven_days_ago = remove_protected(subvol, subvols_seven_days_ago)
        if len(subvols_seven_days_ago) != 0:
            seven_days_ago_culled = generate_weekly_cull_list(subvols_seven_days_ago)
            return_list.append(btrfs_del(subvol.get(location), seven_days_ago_culled, remote,
                               remote_location=subvol.get('remote_location')))

    return return_list


def generate_quarterly_yearly_cull_list(dir_to_cull: list) -> list:
    """Take a list of snapshots from a directory (already reduced to one week or quarter) and cull.

            This culling will return a list with the snapshots to remove for the week or quarter.

            For a perfect set of snapshots, (where the user has been doing one snapshot per hour and \
            doing the daily culling) it should leave behind (remove from list):

            - day7-1800

            .. note:: May end up combining with weekly cull as they essentially do the same thing.

            :param dir_to_cull: A list containing snapshots. Assumes another function\
            has already reduced this list to a list containing only one week or quarter's worth of\
            snapshots.
            :returns: A list containing all the subvolumes to cull.
            """
    sorted_dir_to_cull = sorted(dir_to_cull)
    if len(sorted_dir_to_cull) != 0:
        sorted_dir_to_cull.pop()
    return sorted_dir_to_cull


def remove_protected(subvol: dict, subvol_list_to_pare: list):
    protected_snapshots = subvol.get("remote_protected")
    if protected_snapshots is None:
        protected_snapshots = []
    return [subvol for subvol in subvol_list_to_pare if subvol not in protected_snapshots]


def cull_last_quarter(configuration: dict, remote: bool = False) -> list:
    """Cull the btrfs snapshots from quarter.

    Should leave 1 snapshot per week for 13 weeks.

    :param remote: Are we doing this on the remote system?
    :param configuration: The configuration file.
    :returns: A list containing the results of running the commands.
    """
    last_quarter: list = snapintime.utils.date.quarterly_weeks(datetime.now())
    location: str = "remote_subvol_dir" if remote else "backuplocation"
    return_list = []
    for subvol in configuration.values():
        for week in last_quarter:
            reg_ex_string = ""
            for day in week:
                reg_ex_string = f"{reg_ex_string}({day.strftime('%Y-%m-%d')})|"
            reg_ex_string_minus_final_or = reg_ex_string[:-1]
            weekly_reg_ex = re.compile(reg_ex_string_minus_final_or)
            subvols_this_week = get_subvols_by_date(subvol.get(location), weekly_reg_ex,
                                                    remote, subvol.get("remote_location"))
            if remote:
                subvols_this_week = remove_protected(subvol, subvols_this_week)
            if len(subvols_this_week) != 0:
                this_week_culled = generate_quarterly_yearly_cull_list(subvols_this_week)
                return_list.append(btrfs_del(subvol.get(location), this_week_culled, remote,
                                             remote_location=subvol.get('remote_location')))
    return return_list


def cull_last_year(configuration: dict, remote: bool = False) -> list:
    """Cull the btrfs snapshots from quarter.

    Should leave 1 snapshot per quarter for 4 quarters.

    :param remote: If true, we're doing this on the remote system.
    :param configuration: The configuration file.
    :returns: A list containing the results of running the commands.
    """
    last_year: list = snapintime.utils.date.yearly_quarters(datetime.now())
    return_list = []
    for subvol in configuration.values():
        for quarter in last_year:
            reg_ex_string = ""
            for day in quarter:
                reg_ex_string = f"{reg_ex_string}({day.strftime('%Y-%m-%d')})|"
            reg_ex_string_minus_final_or = reg_ex_string[:-1]
            quarterly_reg_ex = re.compile(reg_ex_string_minus_final_or)
            subvols_this_quarter = get_subvols_by_date(subvol.get("backuplocation"), quarterly_reg_ex)
            if remote:
                subvols_this_quarter = remove_protected(subvol, subvols_this_quarter)
            if len(subvols_this_quarter) != 0:
                this_quarter_culled = generate_quarterly_yearly_cull_list(subvols_this_quarter)
                return_list.append(btrfs_del(subvol.get("backuplocation"), this_quarter_culled, remote,
                                             remote_location=subvol.get('remote_location')))
    return return_list


def print_output(list_of_lists: list):  # pragma: no cover
    for directory in list_of_lists:
        for result in directory:
            log.info(result)


def main():  # pragma: no cover
    our_config = config.import_config()
    three_day_cull_result = cull_three_days_ago(our_config)
    print_output(three_day_cull_result)
    seven_day_cull_result = cull_seven_days_ago(our_config)
    print_output(seven_day_cull_result)
    quarter_cull_result = cull_last_quarter(our_config)
    print_output(quarter_cull_result)
    year_cull_result = cull_last_year(our_config)
    print_output(year_cull_result)


if __name__ == "__main__":  # pragma: no cover
    main()



"""Use btrfs send/receive to send snapshot to a remote computer."""

import os
import subprocess

from snapintime.utils import config as config  # type: ignore

from . import log


def get_remote_subvols(remote_location: str, remote_subvol_dir: str) -> list:
    """Retrieve the remote subvolumes.

    This function assumes user has set up ssh keys for paswordless login.

    :param remote_location: This should be a string like user@computer or\
    user@IPaddress
    :param remote_subvol_dir: This is the directory we will search to get\
    the latest subvolume.
    :returns: A list of the remove subvolumes.
    """
    command = f"ssh {remote_location} ls {remote_subvol_dir}"
    results = subprocess.run(command, capture_output=True, shell=True, check=True, text=True)
    return results.stdout.split('\n')


def get_local_subvols(local_subvol_dir: str) -> list:
    """Grab the subvolumes from the local directory.

    :param local_subvol_dir: The directory containing the local subvolumes.
    :returns: A list of all the local subvolumes
    """
    return os.listdir(path=local_subvol_dir)


def match_subvols(local_subvols: list, remote_subvols: list) -> str:
    """Return the latest remote subvol that also exists on the local system.

    :param local_subvols: A list of the local subvolumes.
    :param remote_subvols: A list of the remote subvolumes.
    :returns: The subvolume.
    """
    sorted_remote = sorted(remote_subvols)
    log.debug(f"{sorted_remote=}")
    candidate = sorted_remote[-1]
    if candidate in local_subvols:
        return candidate
    sorted_remote.pop()
    return match_subvols(local_subvols, sorted_remote)


def btrfs_send_receive(local_subvols: list, remote_subvol: str, backup_location: str,
                       remote_location: str, remote_subvol_dir: str):
    """Run command to send/receive btrfs snapshot.

    :param local_subvols: A list of the local subvolumes to choose from.
    :param remote_subvol: The latest subvolume that is present on both the remote\
    and local systems.
    :param backup_location: The folder prefix for teh local_subvol
    :param remote_location: This should be a string like user@computer or\
    user@IPaddress
    :param remote_subvol_dir: This is the directory we will put the backup into on\
    the remote system.
    :returns: A dictionary with the result of the command.
    """
    sorted_local = sorted(local_subvols)
    command = f"/usr/sbin/btrfs send -p {backup_location}/{remote_subvol} {backup_location}/{sorted_local[-1]} | ssh {remote_location} btrfs receive {remote_subvol_dir}"
    try:
        raw_result = subprocess.run(command, capture_output=True, shell=True, check=True, text=True)
        return {
            "Command": raw_result.args,
            "Return Code": raw_result.returncode,
            "Output": raw_result.stdout,
        }

    except subprocess.SubprocessError as e:
        return {"Command": e.args, "Return Code": e.returncode, "Output": e.stderr}  # type: ignore


def iterate_over_subvolumes(config: dict) -> list:
    """Iterate over all the subvolumes in the config file, then call\
    btrfs_send_receive if the value of remote is "True".

    :param config: The config file, parsed by import_config.
    :returns: A list containing return values from btrfs_send_receive
    """
    return_list = []
    for subvol in config.values():
        if subvol.get("remote") == "True":
            log.info(f"Starting on {subvol.get('subvol')}")
            remote_subvols = get_remote_subvols(subvol.get('remote_location'), subvol.get('remote_subvol_dir'))
            log.debug(f"Remote subvols are: {remote_subvols}")
            local_subvols = get_local_subvols(subvol.get("backuplocation"))
            match = match_subvols(local_subvols, remote_subvols)
            return_list.append(btrfs_send_receive(local_subvols, match, subvol.get('backuplocation'),
                                                  subvol.get('remote_location'), subvol.get('remote_subvol_dir')))
    return return_list


def main():  # pragma: no cover
    log.info("Beginning snapshot remote backups....")
    our_config = config.import_config()
    results = iterate_over_subvolumes(our_config)
    for result in results:
        log.info(f"\nRan {result['Command']} with a return code of {result['Return Code']}")
        log.info(f"Result was: {str(result['Output'])}\n")


if __name__ == "__main__":  # pragma: no cover
    main()



from snapintime.utils import config as config  # type: ignore

from . import culling, log


def main():
    our_config = config.import_config()
    # three day cull doesn't make sense because it's only 1 per day already (at least most days)
    weekly_cull_result = culling.cull_seven_days_ago(our_config, True)
    quarterly_cull_result = culling.cull_last_quarter(our_config, True)
    yearly_cull_result = culling.cull_last_year(our_config, True)
    log.info(weekly_cull_result)
    log.info(quarterly_cull_result)
    log.info(yearly_cull_result)


if __name__ == "__main__":  # pragma: no cover
    log.info("Beginning remote culling...")
    main()



"""Provide date and Time Operations needed by snapintime."""

from datetime import datetime, timedelta


def prior_date(start_date: datetime, day: int = 0) -> datetime:
    """Provide a prior date offset by the variable given in day.

    Unintuitively, positive numbers subtract days.

    :param start_date: The date from which you want to count back or forward.
    :param day: The number of days you want to go back.
    :returns: A datetime object day amount of days in the past.
    """
    delta_days = timedelta(days=day)
    return start_date - delta_days


def many_dates(start_date: datetime, interval_start: int, interval_end: int) -> list:
    """Provide a list of dates within a certain range.

    Used by quarterly culling and yearly culling to determine date range to cull.

    :param start_date: The reference point for the intervals
    :param interval_start: How many days ago you want to start getting dates from.
    :param interval_end: How many days ago you want to stop getting dates from.
    """
    return [
        prior_date(start_date, day)
        for day in range(interval_start, interval_end + 1)
    ]


def quarterly_weeks(start_date: datetime) -> list:
    """Provide a list of 13 weekly date lists.

    :param start_date: Date from which to go back a quarter.
    :returns: A list of lists containing datetime objects. Each sublist represents a week.
    """
    return [
        many_dates(start_date, 90 + number, 96 + number)
        for number in range(0, 90, 7)
    ]


def yearly_quarters(start_date: datetime) -> list:
    """Provide a list of 4 quarterly date lists.

    :param start_date: Date from which to go back a year.
    :returns: A list of lists containing datetime objects. Each sublist represents a quarter."""
    return [
        many_dates(start_date, 365 + number, 455 + number)
        for number in range(0, 275, 91)
    ]



"""Test remote backup functions."""

import subprocess
from unittest import mock

from snapintime import remote_backup


@mock.patch('snapintime.remote_backup.subprocess')
def test_get_remote_latest_subvol(subprocess_mock):
    subprocess_done = subprocess.CompletedProcess(args="args", returncode=0, stdout="2020-02-11\n2020-02-17\n2020-02-20\n2020-02-20p2\n2020-02-25\n2020-03-01\n2020-03-05-2022\n")
    subprocess_mock.run.return_value = subprocess_done
    result = remote_backup.get_remote_subvols("user@remote", "/media/backup")
    assert result == ["2020-02-11", "2020-02-17", "2020-02-20", "2020-02-20p2", "2020-02-25", "2020-03-01", "2020-03-05-2022", '']


@mock.patch('snapintime.remote_backup.os')
def test_get_local_subvols(os_mock):
    mock_return = ["2020-02-11", "2020-02-17", "2020-02-20", "2020-02-20p2", "2020-02-25", "2020-03-01", "2020-03-05-2022"]
    os_mock.listdir.return_value = mock_return
    result = remote_backup.get_local_subvols('/dir')
    assert result == ["2020-02-11", "2020-02-17", "2020-02-20", "2020-02-20p2", "2020-02-25", "2020-03-01", "2020-03-05-2022"]


def test_match_subvols_oldest_remote_is_there():
    """This should not trigger the recursion.

    The final value in the sorted remote list is in the local list.
    """
    local_list = ["A", "B", "C", "D"]
    remote_list = ["A", "B", "C"]
    result = remote_backup.match_subvols(local_list, remote_list)
    assert result == "C"


def test_match_subvols_oldest_remote_not_there():
    """This should trigger the recursion.

    The final value in the sorted remote list is not in the local list.
    """
    local_list = ["A", "B", "D"]
    remote_list = ["A", "B", "C"]
    result = remote_backup.match_subvols(local_list, remote_list)
    assert result == "B"


def test_btrfs_send_receive():
    pass


def test_iterate_configs():
    pass



"""Load the config file."""

import json

import xdgenvpy  # type: ignore


def import_config() -> dict:
    """Import config file.

    :returns: A dictionary containing configs
    :raises: FileNotFoundError
    """
    xdg = xdgenvpy.XDGPedanticPackage('snapintime')
    try:
        with open(f"{xdg.XDG_CONFIG_HOME}/config.json") as file:
            config = json.load(file)
        return config
    except FileNotFoundError:
        print("Could not find config file.")
        raise



"""Read in configuration file and create local snapshots."""

import subprocess
from datetime import datetime

from snapintime import log
from snapintime.utils import config as config  # type: ignore


def get_date_time() -> str:
    """Return the current time, uses system time zone."""
    now: datetime = datetime.now()
    return now.strftime("%Y-%m-%d-%H%M")


def iterate_configs(date_time: str, config: dict) -> list:
    """Iterate over all the subvolumes in the config file, then call\
    create_snapshot.

    :param date_time: The date time that will end up as the btrfs snapshot name
    :param config: The config file, parsed by import_config.
    :returns: A list containing return values from create_snapshot
    """
    return [
        create_snapshot(
            date_time, subvol.get("subvol"), subvol.get("backuplocation")
        )
        for subvol in config.values()
    ]


def create_snapshot(date_suffix: str, subvol: str, backup_location: str):
    """Create a btrfs snapshot.

    :param date_suffix: a datetime object formatted to be the name of the snapshot
    :param subvol: The subvolume to be snapshotted
    :param backup_location: The folder in which to create the snapshot
    """
    command = f"/usr/sbin/btrfs sub snap -r {subvol} {backup_location}/{date_suffix}"
    try:
        raw_result = subprocess.run(command, capture_output=True, shell=True, check=True, text=True)
        return {
            "Command": raw_result.args,
            "Return Code": raw_result.returncode,
            "Output": raw_result.stdout,
        }

    except subprocess.CalledProcessError as e:
        return {
            "Command": e.args[1],
            "Return Code": {e.returncode},
            "Output": str(e.stderr),
        }


def main():  # pragma: no cover
    log.info("Creating local snapshots....")
    date_time_for_backup = get_date_time()
    our_config = config.import_config()
    results = iterate_configs(date_time_for_backup, our_config)
    for result in results:
        log.info(f"Ran: {result['Command']}. Command had a return code of {result['Return Code']}")
        log.info(f"Result was: {str(result['Output'])}")


if __name__ == "__main__":  # pragma: no cover
    main()
